/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.wordsource.WordSourceId;

import com.topcoder.util.compression.CompressionUtility;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.io.Serializable;
import java.io.Writer;

import java.security.DigestInputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;


/**
 * <p>
 * A utility class providing some general operations such as argument checking and stream closing. This class is
 * declared as public for the sub-packages to access. It is not intended for public use however.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
public class Util {
    /**
     * <p>
     * The fully-qualified class name of the Base64 Codec (used by the Compression Utility).
     * </p>
     */
    private static final String BASE64_CODEC_CLASS_NAME = "com.topcoder.util.compression.Base64Codec";

    /**
     * <p>
     * Private constructor to prevent outside instantiation.
     * </p>
     */
    private Util() {
    }

    /**
     * <p>
     * Checks if the document index argument is null.
     * </p>
     *
     * @param documentIndex the document index to check.
     *
     * @throws IllegalArgumentException if documentIndex is null.
     */
    public static void checkDocumentIndex(DocumentIndex documentIndex) {
        if (documentIndex == null) {
            throw new IllegalArgumentException("documentIndex is null");
        }
    }

    /**
     * <p>
     * Checks if the word source id argument is null.
     * </p>
     *
     * @param wordSourceId the word source id to check.
     *
     * @throws IllegalArgumentException if wordSourceId is null.
     */
    public static void checkWordSourceId(WordSourceId wordSourceId) {
        if (wordSourceId == null) {
            throw new IllegalArgumentException("wordSourceId is null");
        }
    }

    /**
     * <p>
     * Checks if the collection index argument is null.
     * </p>
     *
     * @param collectionIndex the collection index to check.
     *
     * @throws IllegalArgumentException if collectionIndex is null.
     */
    public static void checkCollectionIndex(CollectionIndex collectionIndex) {
        if (collectionIndex == null) {
            throw new IllegalArgumentException("collectionIndex is null");
        }
    }

    /**
     * <p>
     * Checks if the collection id argument is null or empty string.
     * </p>
     *
     * @param collectionId the collection id to check.
     *
     * @throws IllegalArgumentException if collectionId is null or empty string.
     */
    public static void checkCollectionId(String collectionId) {
        if (collectionId == null) {
            throw new IllegalArgumentException("collectionId is null");
        }
        if (collectionId.trim().length() == 0) {
            throw new IllegalArgumentException("collectionId is empty string");
        }
    }

    /**
     * <p>
     * Try to close the InputStream object if it is not null. Returns silently if any I/O error occurs.
     * </p>
     *
     * @param in the InputStream object.
     */
    public static void close(InputStream in) {
        if (in != null) {
            try {
                in.close();
            } catch (IOException ex) {
                // ignore any I/O error
            }
        }
    }

    /**
     * <p>
     * Try to close the OutputStream object if it is not null. Returns silently if any I/O error occurs.
     * </p>
     *
     * @param out the OutputStream object.
     */
    public static void close(OutputStream out) {
        if (out != null) {
            try {
                out.close();
            } catch (IOException ex) {
                // ignore any I/O error
            }
        }
    }

    /**
     * Try to close the reader object if it is not null. Returns silently if any I/O error occurs.
     *
     * @param reader the reader object.
     */
    public static void close(Reader reader) {
        if (reader != null) {
            try {
                reader.close();
            } catch (IOException e) {
                // ignore any I/O error
            }
        }
    }

    /**
     * Try to close the writer object if it is not null. Returns silently if any I/O error occurs.
     *
     * @param writer the writer object.
     */
    public static void close(Writer writer) {
        if (writer != null) {
            try {
                writer.close();
            } catch (IOException e) {
                // ignore any I/O error
            }
        }
    }

    /**
     * <p>
     * Serializes the object, then encodes the byte data to Base64 string representation.
     * </p>
     *
     * @param obj the object to compress.
     *
     * @return the Base64 string representation.
     *
     * @throws IndexPersistenceException if any error occurs during compression.
     */
    public static String compress(Object obj) throws IndexPersistenceException {
        ByteArrayOutputStream out = new ByteArrayOutputStream();

        // serialize the object
        byte[] data = serialize(obj);

        // compress the byte data using Base64 algorithm
        try {
            CompressionUtility cu = new CompressionUtility(BASE64_CODEC_CLASS_NAME, out);

            cu.compress(new ByteArrayInputStream(data));
        } catch (Exception e) {
            throw new IndexPersistenceException("Some error occurs in compression", e);
        }

        return new String(out.toByteArray());
    }

    /**
     * <p>
     * Decodes the Base64 string representation, then deserializes the byte data back to the original object.
     * </p>
     *
     * @param input the Base64 string representation.
     *
     * @return the decompressed object.
     *
     * @throws IndexPersistenceException if any error occurs during decompression.
     */
    public static Object decompress(String input) throws IndexPersistenceException {
        ByteArrayOutputStream out = new ByteArrayOutputStream();

        // decode the string using Base64 algorithm
        try {
            CompressionUtility cu = new CompressionUtility(BASE64_CODEC_CLASS_NAME, out);

            cu.decompress(new StringBuffer(input));
        } catch (Exception e) {
            throw new IndexPersistenceException("Some error occurs in decompression", e);
        }

        // deserialize the byte data
        return deserialize(out.toByteArray());
    }

    /**
     * <p>
     * Serializes the specified object into an array of bytes.
     * </p>
     *
     * @param obj the object to serialize.
     *
     * @return an array of bytes.
     *
     * @throws IndexPersistenceException if the object cannot be serialized.
     */
    public static byte[] serialize(Object obj) throws IndexPersistenceException {
        // check if the object is serializable
        if (!(obj instanceof Serializable)) {
            throw new IndexPersistenceException("obj is not of type Serializable");
        }

        ByteArrayOutputStream outStream = new ByteArrayOutputStream();
        ObjectOutputStream objStream = null;

        try {
            // serialize the object
            objStream = new ObjectOutputStream(outStream);
            objStream.writeObject(obj);

            // get the serialized form in byte array
            return outStream.toByteArray();
        } catch (IOException e) {
            throw new IndexPersistenceException("Fails to serialize object", e);
        } finally {
            close(objStream);
        }
    }

    /**
     * <p>
     * Serializes the specified word source id into an array of bytes. This method is needed because WordSourceId is
     * not of Serializable type.
     * </p>
     *
     * @param wordSourceId the word source id to serialize.
     *
     * @return an array of bytes.
     *
     * @throws IndexPersistenceException if the word source id cannot be serialized.
     */
    public static byte[] serialize(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        ByteArrayOutputStream out = new ByteArrayOutputStream();

        try {
            out.write(serialize(wordSourceId.getSourceIdentity()));
            out.write(serialize(wordSourceId.getSourceClassName()));
            out.write(serialize(wordSourceId.getSourceLocale()));
            out.write(serialize(wordSourceId.getDelimiters()));
            return out.toByteArray();
        } catch (IOException e) {
            // should not happen
            return null;
        }
    }

    /**
     * <p>
     * Deserializes the specified array of bytes into an object.
     * </p>
     *
     * @param data the array of bytes to deserialize.
     *
     * @return the deserialized object.
     *
     * @throws IndexPersistenceException if the object cannot be deserialized.
     */
    public static Object deserialize(byte[] data) throws IndexPersistenceException {
        ByteArrayInputStream inStream = new ByteArrayInputStream(data);
        ObjectInputStream objStream = null;

        try {
            // deserialize the object
            objStream = new ObjectInputStream(inStream);
            return objStream.readObject();
        } catch (IOException e) {
            throw new IndexPersistenceException("Fails to deserialize object", e);
        } catch (ClassNotFoundException e) {
            throw new IndexPersistenceException("Unable to locate the class", e);
        } finally {
            close(objStream);
        }
    }

    /**
     * <p>
     * Gets the unique document ids for the given WordSourceId instances.
     * </p>
     *
     * @param wordSourceIds the WordSourceId instances.
     *
     * @return the unique document ids.
     *
     * @throws IndexPersistenceException if any wordSourceId cannot be serialized; or generating the hash value fails.
     */
    public static Set getDocumentIds(Set wordSourceIds)
        throws IndexPersistenceException {
        Set documentIds = new HashSet();

        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            documentIds.add(getDocumentId((WordSourceId) i.next()));
        }
        return documentIds;
    }

    /**
     * <p>
     * Gets the unique document id for the given WordSourceId instance.
     * </p>
     *
     * @param wordSourceId the WordSourceId instance.
     *
     * @return the unique document id.
     *
     * @throws IndexPersistenceException if the wordSourceId cannot be serialized; or generating the hash value fails.
     */
    public static String getDocumentId(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        // serialize the word source id
        byte[] data = serialize(wordSourceId);

        DigestInputStream in = null;

        try {
            // use the MD5 algorithm to generate the hash
            MessageDigest md = MessageDigest.getInstance("MD5");

            in = new DigestInputStream(new ByteArrayInputStream(data), md);
            while (in.read() != -1) {
                // do nothing to update message digest
            }

            // return the hex string representation of the hash value
            return toHexString(md.digest());
        } catch (NoSuchAlgorithmException e) {
            throw new IndexPersistenceException("Digest algorithm is not supported", e);
        } catch (IOException e) {
            throw new IndexPersistenceException("Some I/O error occurs", e);
        } finally {
            close(in);
        }
    }

    /**
     * <p>
     * Converts a byte array to a hex string, with leading zeros.
     * </p>
     *
     * @param data the byte array to convert.
     *
     * @return the hex string representation.
     */
    public static String toHexString(byte[] data) {
        StringBuffer sb = new StringBuffer();

        for (int i = 0; i < data.length; i++) {
            int byteValue = (0x000000ff & data[i]) | 0xffffff00;

            sb.append(Integer.toHexString(byteValue).substring(6));
        }
        return sb.toString();
    }

    /**
     * <p>
     * Converts a character to a hex string (in Unicode), with leading zeros.
     * </p>
     *
     * @param ch the character to convert.
     *
     * @return the hex string representation.
     */
    public static String toHexString(char ch) {
        byte[] data = new byte[2];

        data[0] = (byte) (ch >>> 8);
        data[1] = (byte) (ch & 0x00ff);
        return toHexString(data);
    }
}
