/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.db.connectionfactory.DBConnectionException;
import com.topcoder.db.connectionfactory.DBConnectionFactory;
import com.topcoder.db.connectionfactory.DBConnectionFactoryImpl;

import com.topcoder.document.index.persistence.IndexPersistence;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.ConfigUtil;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;

import com.topcoder.util.config.Property;

import java.lang.reflect.InvocationTargetException;

import java.sql.Connection;
import java.sql.SQLException;


/**
 * <p>
 * This is an abstraction of a typical database driven persistence contract. It provides the methods for getting a
 * connection using the DBConnectionFactory component as well as controlling transactions.
 * </p>
 *
 * <p>
 * It supports the loading of connection-related properties from configuration file. Here is a sample:
 * <pre>
 * &lt;?xml version="1.0"?&gt;
 * &lt;CMConfig&gt;
 *   &lt;Config name="com.topcoder.document.index.persistence.impl.db"&gt;
 *     &lt;!-- the class name of the connection factory (optional) --&gt;
 *     &lt;Property name="ConnectionFactoryClassName"&gt;
 *       &lt;Value&gt;com.topcoder.db.connectionfactory.DBConnectionFactoryImpl&lt;/Value&gt;
 *     &lt;/Property&gt;
 *     &lt;!-- the namespace to pass to the connection factory (required) --&gt;
 *     &lt;Property name="ConnectionFactoryNamespace"&gt;
 *       &lt;Value&gt;com.topcoder.db.connectionfactory&lt;/Value&gt;
 *     &lt;/Property&gt;
 *     &lt;!-- the name of the connection to be used (optional) --&gt;
 *     &lt;Property name="ConnectionName"&gt;
 *       &lt;Value&gt;Informix_Connection_Producer&lt;/Value&gt;
 *     &lt;/Property&gt;
 *   &lt;/Config&gt;
 * &lt;/CMConfig&gt;
 * </pre>
 * </p>
 *
 * <p>
 * This class is not thread-safe. Implementations are not required to be thread-safe.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
public abstract class AbstractDBIndexPersistence implements IndexPersistence {
    /**
     * <p>
     * Constant for the property specifying the class name of the DB Connection Factory.
     * </p>
     */
    private static final String CONN_FACTORY_CLASSNAME_PROPERTY = "ConnectionFactoryClassName";

    /**
     * <p>
     * Constant for the property specifying the namespace of the DB Connection Factory configuration file.
     * </p>
     */
    private static final String CONN_FACTORY_NAMESPACE_PROPERTY = "ConnectionFactoryNamespace";

    /**
     * <p>
     * Constant for the property specifying the connection producer name.
     * </p>
     */
    private static final String CONN_NAME_PROPERTY = "ConnectionName";

    /**
     * <p>
     * This represents the connection factory to obtain a pre-configured connection for database access. This is
     * initialized in the constructor and cannot be changed. Cannot be null.
     * </p>
     *
     * <p></p>
     */
    private final DBConnectionFactory connectionFactory;

    /**
     * <p>
     * This represents a connection name that is used to fetch a connection from the connection factory. This is
     * initialized in the constructor and cannot be changed. Can be null or an empty string, upon which it will try to
     * use the default connection.
     * </p>
     */
    private final String connectionName;

    /**
     * <p>
     * A constructor that loads the connection-related properties from the specified namespace.
     * </p>
     *
     * @param namespace The namespace for connection-related properties.
     *
     * @throws IllegalArgumentException if namespace is null or empty string.
     * @throws PersistenceConfigurationException if there are configuration issues encountered.
     */
    protected AbstractDBIndexPersistence(String namespace)
        throws PersistenceConfigurationException {
        if (namespace == null) {
            throw new IllegalArgumentException("namespace is null");
        }
        if (namespace.trim().length() == 0) {
            throw new IllegalArgumentException("namespace is empty string");
        }

        Property root = ConfigUtil.getRootProperty(namespace);

        // read the connection name property
        connectionName = root.getValue(CONN_NAME_PROPERTY);

        // read the connection factory properties
        String connFactoryClassName = ConfigUtil.getOptionalPropertyValue(root, CONN_FACTORY_CLASSNAME_PROPERTY);
        String connFactoryNamespace = ConfigUtil.getPropertyValue(root, CONN_FACTORY_NAMESPACE_PROPERTY);

        Class connFactoryClass = null;

        if (connFactoryClassName == null) {
            // if the connection factory class name is not specified, use the default class
            connFactoryClass = DBConnectionFactoryImpl.class;
        } else {
            // otherwise locate the connection factory class
            try {
                connFactoryClass = Class.forName(connFactoryClassName);
            } catch (ClassNotFoundException e) {
                throw new PersistenceConfigurationException("Unable to locate the class", e);
            }
            if (!DBConnectionFactory.class.isAssignableFrom(connFactoryClass)) {
                throw new PersistenceConfigurationException("Class " + connFactoryClassName
                    + " does not implement DBConnectionFactory");
            }
        }

        // prepare the constructor parameters
        Class[] parameterTypes = new Class[] {String.class};
        Object[] parameters = new String[] {connFactoryNamespace};

        // instantiate the connection factory class
        try {
            connectionFactory = (DBConnectionFactory)
                connFactoryClass.getConstructor(parameterTypes).newInstance(parameters);
        } catch (NoSuchMethodException e) {
            throw new PersistenceConfigurationException("Unable to locate the constructor", e);
        } catch (InstantiationException e) {
            throw new PersistenceConfigurationException("Fails to create the instance", e);
        } catch (IllegalAccessException e) {
            throw new PersistenceConfigurationException("Fails to create the instance", e);
        } catch (InvocationTargetException e) {
            throw new PersistenceConfigurationException("Fails to create the instance", e);
        }
    }

    /**
     * <p>
     * A constructor that accepts the given connection factory and connection name. If connectionName is null or empty
     * string, the default connection will be used.
     * </p>
     *
     * @param connectionFactory the connection factory used to generate connections.
     * @param connectionName the name of the connection used for persistence operations.
     *
     * @throws IllegalArgumentException if connectionFactory is null.
     */
    protected AbstractDBIndexPersistence(DBConnectionFactory connectionFactory, String connectionName) {
        if (connectionFactory == null) {
            throw new IllegalArgumentException("connectionFactory is null");
        }
        this.connectionFactory = connectionFactory;
        this.connectionName = connectionName;
    }

    /**
     * <p>
     * Returns a new connection to be used for persistence. The returned connection has the auto-commit mode disabled.
     * The commitTransaction() method must be called explicitly in order to commit changes; otherwise, database
     * changes will not be saved.
     * </p>
     *
     * @return the new connection.
     *
     * @throws IndexPersistenceException if it is unable to obtain a connection, or if the connection does not support
     *         transactions.
     */
    protected Connection getConnection() throws IndexPersistenceException {
        Connection connection = null;

        try {
            if ((connectionName == null) || (connectionName.trim().length() == 0)) {
                // if connection name is not specified, use the default connection
                connection = connectionFactory.createConnection();
            } else {
                // otherwise get the connection with the connection name
                connection = connectionFactory.createConnection(connectionName);
            }

            // disable the auto-commit mode
            connection.setAutoCommit(false);
            return connection;
        } catch (DBConnectionException e) {
            throw new IndexPersistenceException("Fails to obtain a connection", e);
        } catch (SQLException e) {
            DBUtil.close(connection);
            throw new IndexPersistenceException("Fails to start the transaction", e);
        }
    }

    /**
     * <p>
     * Starts the transaction on the given connection. Does nothing here (except argument checking) since the
     * connection returned from getConnection() method already has the auto-commit mode disabled.
     * </p>
     *
     * @param connection the connection taking part in the transaction.
     *
     * @throws IllegalArgumentException if connection is null.
     * @throws IndexPersistenceException if any error occurs during this process, or if the connection does not support
     *         transactions.
     */
    protected void beginTransaction(Connection connection)
        throws IndexPersistenceException {
        checkConnection(connection);
    }

    /**
     * <p>
     * Commits the transaction on the given connection. The connection is closed after calling this method.
     * </p>
     *
     * @param connection the connection taking part in the transaction.
     *
     * @throws IllegalArgumentException if connection is null.
     * @throws IndexPersistenceException if any error occurs during this process, or if the connection does not support
     *         transactions.
     */
    protected void commitTransaction(Connection connection)
        throws IndexPersistenceException {
        checkConnection(connection);

        try {
            connection.commit();
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to commit the transaction", e);
        } finally {
            DBUtil.close(connection);
        }
    }

    /**
     * <p>
     * Rolls back the transaction on the given connection. The connection is closed after calling this method.
     * </p>
     *
     * @param connection the connection taking part in the transaction.
     *
     * @throws IllegalArgumentException if connection is null.
     * @throws IndexPersistenceException if any error occurs during this process, or if the connection does not support
     *         transactions.
     */
    protected void rollbackTransaction(Connection connection)
        throws IndexPersistenceException {
        checkConnection(connection);

        try {
            connection.rollback();
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to rollback the transaction", e);
        } finally {
            DBUtil.close(connection);
        }
    }

    /**
     * <p>
     * Checks if the connection object argument is null.
     * </p>
     *
     * @param connection the connection object to check.
     *
     * @throws IllegalArgumentException if connection is null.
     */
    private void checkConnection(Connection connection) {
        if (connection == null) {
            throw new IllegalArgumentException("connection is null");
        }
    }
}
