/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import java.util.ArrayList;
import java.util.List;


/**
 * <p>
 * A utility class providing some common operations related to database. This class and all its methods are
 * package-private, so that they can only be used by classes in the same package.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
class DBUtil {
    /**
     * <p>
     * Private constructor to prevent outside instantiation.
     * </p>
     */
    private DBUtil() {
    }

    /**
     * <p>
     * Try to close the connection object. Returns silently if any SQL error occurs.
     * </p>
     *
     * @param con the connection object.
     */
    static void close(Connection con) {
        try {
            con.close();
        } catch (SQLException e) {
            // ignore SQL error
        }
    }

    /**
     * <p>
     * Try to close the prepared statement object if it is not null. Returns silently if any SQL error occurs.
     * </p>
     *
     * @param pstmt the prepared statement object.
     */
    static void close(PreparedStatement pstmt) {
        if (pstmt != null) {
            try {
                pstmt.close();
            } catch (SQLException e) {
                // ignore SQL error
            }
        }
    }

    /**
     * <p>
     * Try to close the result set object if it is not null. Returns silently if any SQL error occurs.
     * </p>
     *
     * @param rs the result set object.
     */
    static void close(ResultSet rs) {
        if (rs != null) {
            try {
                rs.close();
            } catch (SQLException e) {
                // ignore SQL error
            }
        }
    }

    /**
     * <p>
     * A convenient method to execute the SQL statement with one string parameter. The SQL statement should be an
     * INSERT, DELETE or UPDATE statement.
     * </p>
     *
     * <p>
     * This method throws SQLException so that the caller can wrap this exception with specific message.
     * </p>
     *
     * @param connection the connection to the database.
     * @param sql the SQL statement to execute
     * @param param1 the string parameter
     *
     * @return whether any row was updated
     *
     * @throws SQLException if any database error occurs
     */
    static boolean execute(Connection connection, String sql, String param1)
        throws SQLException {
        PreparedStatement pstmt = null;

        try {
            // set one string parameter
            pstmt = connection.prepareStatement(sql);
            pstmt.setString(1, param1);
            return pstmt.executeUpdate() > 0;
        } finally {
            close(pstmt);
        }
    }

    /**
     * <p>
     * A convenient method to execute the SQL statement with two string parameters. The first string parameter is
     * fixed, while the second string parameter changes for each execution. The SQL statement should be an INSERT,
     * DELETE or UPDATE statement.
     * </p>
     *
     * <p>
     * This method throws SQLException so that the caller can wrap this exception with specific message.
     * </p>
     *
     * @param connection the connection to the database.
     * @param sql the SQL statement to execute
     * @param param1 the first string parameter
     * @param param2 the second string parameters
     *
     * @throws SQLException if any database error occurs
     */
    static void execute(Connection connection, String sql, String param1, String[] param2)
        throws SQLException {
        execute(connection, sql, param1, param2, false);
    }

    /**
     * <p>
     * A convenient method to execute the SQL statement with two string parameters. The first string parameter is
     * fixed, while the second string parameter changes for each execution. The hasParam3 flag indicates whether there
     * is a third integer parameter specifying the index position of the second string parameter. The SQL statement
     * should be an INSERT, DELETE or UPDATE statement.
     * </p>
     *
     * <p>
     * This method throws SQLException so that the caller can wrap this exception with specific message.
     * </p>
     *
     * @param connection the connection to the database.
     * @param sql the SQL statement to execute
     * @param param1 the first string parameter
     * @param param2 the second string parameters
     * @param hasParam3 whether there is a third integer parameter
     *
     * @throws SQLException if any database error occurs
     */
    static void execute(Connection connection, String sql, String param1, String[] param2, boolean hasParam3)
        throws SQLException {
        PreparedStatement pstmt = null;

        try {
            // set first string parameter
            pstmt = connection.prepareStatement(sql);
            pstmt.setString(1, param1);
            for (int i = 0; i < param2.length; i++) {
                // set second string parameter
                pstmt.setString(2, param2[i]);
                if (hasParam3) {
                    // set third integer parameter
                    pstmt.setInt(3, i);
                }
                pstmt.executeUpdate();
            }
        } finally {
            close(pstmt);
        }
    }

    /**
     * <p>
     * A convenient method to check the existence of a row in the database by executing the SQL statement with one
     * string parameter. The SQL statement should be a SELECT statement.
     * </p>
     *
     * <p>
     * This method throws SQLException so that the caller can wrap this exception with specific message.
     * </p>
     *
     * @param connection the connection to the database.
     * @param sql the SQL statement to execute
     * @param param1 the string parameter
     *
     * @return whether the row exists or not
     *
     * @throws SQLException if any database error occurs
     */
    static boolean exist(Connection connection, String sql, String param1)
        throws SQLException {
        PreparedStatement pstmt = null;
        ResultSet rs = null;

        try {
            // set one string parameter
            pstmt = connection.prepareStatement(sql);
            pstmt.setString(1, param1);
            rs = pstmt.executeQuery();
            return rs.next();
        } finally {
            close(rs);
            close(pstmt);
        }
    }

    /**
     * <p>
     * A convenient method to get the strings returned from the result set. The SQL statement should be a SELECT
     * statement on a single column.
     * </p>
     *
     * <p>
     * This method throws SQLException so that the caller can wrap this exception with specific message.
     * </p>
     *
     * @param rs the result set containing the data
     *
     * @return a list of Strings
     *
     * @throws SQLException if any database error occurs
     */
    static List getStrings(ResultSet rs) throws SQLException {
        List result = new ArrayList();

        while (rs.next()) {
            result.add(rs.getString(1));
        }
        return result;
    }

    /**
     * <p>
     * A convenient method to get the integers returned from the result set. The SQL statement should be a SELECT
     * statement on a single column.
     * </p>
     *
     * <p>
     * This method throws SQLException so that the caller can wrap this exception with specific message.
     * </p>
     *
     * @param rs the result set containing the data
     *
     * @return a list of Integers
     *
     * @throws SQLException if any database error occurs
     */
    static List getIntegers(ResultSet rs) throws SQLException {
        List result = new ArrayList();

        while (rs.next()) {
            result.add(new Integer(rs.getInt(1)));
        }
        return result;
    }
}
