/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.Util;
import com.topcoder.document.index.wordsource.WordSourceId;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import java.util.HashSet;
import java.util.Locale;
import java.util.Set;


/**
 * <p>
 * A class providing some common operations related to database persistence of document index. This class assumes that
 * the DOCUMENT and DELIMITER tables are common to the database persistence.
 * </p>
 *
 * <p>
 * This class and all its methods are package-private, so that they can only be used by classes in the same package. We
 * do not put the methods inside AbstractDBIndexPersistence because it is not required to have such table structures.
 * All methods assume valid arguments.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
class DocumentIndexPersistence {
    /**
     * <p>
     * The SQL statement to insert a document - the use count is initially set to 0.
     * </p>
     */
    private static final String SQL_INS_DOCUMENT = "INSERT INTO DOCUMENT VALUES (?, ?, ?, ?, 0)";

    /**
     * <p>
     * The SQL statement to insert a delimiter.
     * </p>
     */
    private static final String SQL_INS_DELIMITER = "INSERT INTO DELIMITER VALUES (?, ?, ?)";

    /**
     * <p>
     * The SQL statement to delete a document.
     * </p>
     */
    private static final String SQL_DEL_DOCUMENT = "DELETE FROM DOCUMENT WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * The SQL statement to delete the delimiters belonging to a document.
     * </p>
     */
    private static final String SQL_DEL_DELIMITER = "DELETE FROM DELIMITER WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * The SQL statement to select an unused document - the use count must be 0.
     * </p>
     */
    private static final String SQL_SEL_UNUSED_DOCUMENT = "SELECT DOCUMENT_ID FROM DOCUMENT WHERE DOCUMENT_ID = ? "
        + "AND USE_COUNT = 0";

    /**
     * <p>
     * The SQL statement to select a document.
     * </p>
     */
    private static final String SQL_SEL_DOCUMENT = "SELECT DOCUMENT_ID FROM DOCUMENT WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * The SQL statement to select the delimiters belonging to a document.
     * </p>
     */
    private static final String SQL_SEL_DELIMITERS = "SELECT DELIMITER_ENTRY FROM DELIMITER WHERE DOCUMENT_ID = ? "
        + "ORDER BY DELIMITER_SORT_ORDER";

    /**
     * <p>
     * The SQL statement to select all documents.
     * </p>
     */
    private static final String SQL_SEL_ALL_DOCUMENTS = "SELECT DOCUMENT_ID, LOCALE, SOURCE_IDENTITY, CLASS_NAME "
        + "FROM DOCUMENT";

    /**
     * <p>
     * The SQL statement to increase the use count belonging to a document by 1.
     * </p>
     */
    private static final String SQL_INCREASE_USE_COUNT = "UPDATE DOCUMENT SET USE_COUNT = USE_COUNT + 1 "
        + "WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * The SQL statement to decrease the use count belonging to a document by 1.
     * </p>
     */
    private static final String SQL_DECREASE_USE_COUNT = "UPDATE DOCUMENT SET USE_COUNT = USE_COUNT - 1 "
        + "WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * The SQL statement to select the use count belonging to a document.
     * </p>
     */
    private static final String SQL_SEL_USE_COUNT = "SELECT USE_COUNT FROM DOCUMENT WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * The actual database persistence implementation to access the database. This is initialized in the constructor
     * and cannot be changed. Cannot be null.
     * </p>
     */
    private final AbstractDBIndexPersistence persistence;

    /**
     * <p>
     * Constructor that accepts the actual database persistence implementation.
     * </p>
     *
     * @param persistence the actual database persistence implementation.
     */
    DocumentIndexPersistence(AbstractDBIndexPersistence persistence) {
        this.persistence = persistence;
    }

    /**
     * <p>
     * Adds the given document index to the persistence. The initial collection reference count should be 0. When the
     * document index exists in the persistence, it fails.
     * </p>
     *
     * <p>
     * This method adds records to the DOCUMENT and DELIMITER tables.
     * </p>
     *
     * <p>
     * This method does not deal with transaction. The transaction is expected to be started by the caller so that it
     * can be committed or rolled back in one step, in case there are multiple SQL statements grouped into the
     * transaction. This method assumes valid arguments.
     * </p>
     *
     * @param connection the connection to the database.
     * @param documentId the unique id of the document whose index is persisted.
     * @param documentIndex the document index to be persisted.
     *
     * @throws IndexPersistenceException if persisting the document index fails; or the document index exists in the
     *         persistence.
     */
    void addDocumentIndex(Connection connection, String documentId, DocumentIndex documentIndex)
        throws IndexPersistenceException {
        WordSourceId wordSourceId = documentIndex.getWordSourceId();

        PreparedStatement pstmt = null;

        try {
            // add the document to the DOCUMENT table
            // we use setBytes() for source identity since it won't be too long, and it supports BLOB as well:
            // http://publib.boulder.ibm.com/infocenter/idshelp/v10/topic/com.ibm.jdbc.doc/jdbc218.htm
            pstmt = connection.prepareStatement(SQL_INS_DOCUMENT);
            pstmt.setString(1, documentId);
            pstmt.setString(2, wordSourceId.getSourceLocale().toString());
            pstmt.setBytes(3, Util.serialize(wordSourceId.getSourceIdentity()));
            pstmt.setString(4, wordSourceId.getSourceClassName());
            pstmt.executeUpdate();
            pstmt.close();

            // add the delimiters to the DELIMITER table
            DBUtil.execute(connection, SQL_INS_DELIMITER, documentId, wordSourceId.getDelimiters(), true);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to add document index", e);
        } finally {
            DBUtil.close(pstmt);
        }
    }

    /**
     * <p>
     * Removes a document index from the persistence. The document index can be deleted only when there is no more
     * collection containing the document index. The unique document id is given. When the document is not persisted,
     * it fails.
     * </p>
     *
     * <p>
     * This method deletes records from the DOCUMENT and DELIMITER tables. Normally the caller should have already
     * checked whether the document index can be deleted. It should exist and have a use count equal to 0.
     * </p>
     *
     * <p>
     * This method does not deal with transaction. The transaction is expected to be started by the caller so that it
     * can be committed or rolled back in one step, in case there are multiple SQL statements grouped into the
     * transaction. This method assumes valid arguments.
     * </p>
     *
     * @param connection the connection to the database.
     * @param documentId the unique id of the document whose index is deleted.
     *
     * @throws IndexPersistenceException if deleting the document index fails.
     */
    void removeDocumentIndex(Connection connection, String documentId)
        throws IndexPersistenceException {
        try {
            // remove the delimiters from the DELIMITER table
            DBUtil.execute(connection, SQL_DEL_DELIMITER, documentId);

            // remove the document from the DOCUMENT table
            DBUtil.execute(connection, SQL_DEL_DOCUMENT, documentId);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to remove document index", e);
        }
    }

    /**
     * <p>
     * Returns whether the document index exists and is unused (a use count of 0) in the persistence. The unique
     * document id is given. This method should be called prior to the removal of document index.
     * </p>
     *
     * <p>
     * This method retrieves a record from the DOCUMENT table.
     * </p>
     *
     * @param connection the connection to the database.
     * @param documentId the unique id of the document whose index is checked.
     *
     * @return whether the unused document index exists.
     *
     * @throws IndexPersistenceException if checking the unused document index fails.
     */
    boolean existUnusedDocumentIndex(Connection connection, String documentId)
        throws IndexPersistenceException {
        try {
            // check the unused document in the DOCUMENT table
            return DBUtil.exist(connection, SQL_SEL_UNUSED_DOCUMENT, documentId);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to check unused document index", e);
        }
    }

    /**
     * <p>
     * Returns whether the document index exists in the persistence. The unique document id is given. This method
     * should be called prior to the retrieval of document index.
     * </p>
     *
     * <p>
     * This method retrieves a record from the DOCUMENT table.
     * </p>
     *
     * @param connection the connection to the database.
     * @param documentId the unique id of the document whose index is checked.
     *
     * @return whether the document index exists.
     *
     * @throws IndexPersistenceException if checking the document index fails.
     */
    boolean existDocumentIndex(Connection connection, String documentId)
        throws IndexPersistenceException {
        try {
            // check the document in the DOCUMENT table
            return DBUtil.exist(connection, SQL_SEL_DOCUMENT, documentId);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to check document index", e);
        }
    }

    /**
     * <p>
     * Gets a set of all IDs of documents whose indexes are persisted in the persistence.
     * </p>
     *
     * <p>
     * This method retrieves records from the DOCUMENT and DELIMITER tables.
     * </p>
     *
     * @return a set of all IDs of persisted document indexes.
     *
     * @throws IndexPersistenceException if retrieving the document IDs fails.
     */
    Set getIndexedDocuments() throws IndexPersistenceException {
        Connection connection = persistence.getConnection();
        PreparedStatement pstmt = null;

        try {
            // get the IDs of all document indexes
            pstmt = connection.prepareStatement(SQL_SEL_ALL_DOCUMENTS);
            return getWordSourceIds(pstmt);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to get indexed documents", e);
        } finally {
            DBUtil.close(pstmt);
            DBUtil.close(connection);
        }
    }

    /**
     * <p>
     * Gets the IDs of the document index selected by the given prepared statement. The prepared statement should
     * select the columns DOCUMENT_ID, LOCALE, SOURCE_IDENTITY and CLASS_NAME from the DOCUMENT table. It is expected
     * to be closed by the caller.
     * </p>
     *
     * <p>
     * This method retrieves records from the DOCUMENT and DELIMITER tables.
     * </p>
     *
     * @param pstmt the prepared statement to execute.
     *
     * @return a set of WordSourceId instances.
     *
     * @throws IndexPersistenceException if retrieving the IDs of the document index fails.
     */
    static Set getWordSourceIds(PreparedStatement pstmt)
        throws IndexPersistenceException {
        PreparedStatement pstmt2 = null;
        ResultSet rs = null;
        ResultSet rs2 = null;

        try {
            Connection connection = pstmt.getConnection();

            // get the documents from the DOCUMENT table
            pstmt2 = connection.prepareStatement(SQL_SEL_DELIMITERS);
            rs = pstmt.executeQuery();

            Set wordSourceIds = new HashSet();

            while (rs.next()) {
                // extract the fields of word source id
                // we use getBytes() for source identity since it won't be too long, and it supports BLOB as well:
                // http://publib.boulder.ibm.com/infocenter/idshelp/v10/topic/com.ibm.jdbc.doc/jdbc218.htm
                String documentId = rs.getString(1);
                Locale locale = parseLocale(rs.getString(2));
                Object sourceIdentity = Util.deserialize(rs.getBytes(3));
                String className = rs.getString(4);

                // get the delimiters from the DELIMITER table
                // we open another result set here since the number of delimiters is not large,
                // and the WordSourceId instance can be built right away
                pstmt2.setString(1, documentId);
                rs2 = pstmt2.executeQuery();

                // add the WordSourceId instance to the list
                wordSourceIds.add(new WordSourceId(sourceIdentity, className,
                        (String[]) DBUtil.getStrings(rs2).toArray(new String[0]), locale));

                // free up the resources immediately
                rs2.close();
            }
            return wordSourceIds;
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to get word source ids");
        } catch (IllegalArgumentException e) {
            // wrap the IllegalArgumentException in case the data is corrupted such that
            // the WordSourceId instance cannot be built
            throw new IndexPersistenceException("Unable to create WordSourceId instance", e);
        } finally {
            DBUtil.close(rs);
            DBUtil.close(rs2);
            DBUtil.close(pstmt2);
        }
    }

    /**
     * <p>
     * Increases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * <p>
     * This method updates a record in the DOCUMENT table.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is increased by 1.
     *
     * @throws IndexPersistenceException if increasing the collection reference count fails; or the document index is
     *         not persisted.
     */
    void increaseDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        updateDocumentUseCount(wordSourceId, SQL_INCREASE_USE_COUNT);
    }

    /**
     * <p>
     * Decreases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * <p>
     * This method updates a record in the DOCUMENT table.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is decreased by 1.
     *
     * @throws IndexPersistenceException if decreasing the collection reference count fails; or the document index is
     *         not persisted.
     */
    void decreaseDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        updateDocumentUseCount(wordSourceId, SQL_DECREASE_USE_COUNT);
    }

    /**
     * <p>
     * Updates the collection reference count of the document index. The document ID is given. When the document index
     * is not persisted, it fails.
     * </p>
     *
     * <p>
     * This method updates a record in the DOCUMENT table.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is updated.
     * @param sql the sql query to update the collection reference count.
     *
     * @throws IndexPersistenceException if updating the collection reference count fails; or the document index is not
     *         persisted.
     */
    void updateDocumentUseCount(WordSourceId wordSourceId, String sql)
        throws IndexPersistenceException {
        // get the document id from the word source id
        String documentId = Util.getDocumentId(wordSourceId);

        Connection connection = persistence.getConnection();

        try {
            // start the transaction
            persistence.beginTransaction(connection);

            // update the use count in the DOCUMENT table
            if (!DBUtil.execute(connection, sql, documentId)) {
                throw new IndexPersistenceException("Document index does not exist");
            }

            // commit the transaction
            persistence.commitTransaction(connection);
        } catch (IndexPersistenceException e) {
            persistence.rollbackTransaction(connection);
            throw e;
        } catch (SQLException e) {
            persistence.rollbackTransaction(connection);
            throw new IndexPersistenceException("Fails to update use count", e);
        } finally {
            DBUtil.close(connection);
        }
    }

    /**
     * <p>
     * Gets the collection reference count of the given document index. When the document index is not persisted, it
     * fails.
     * </p>
     *
     * <p>
     * This method retrieves a record from the DOCUMENT table.
     * </p>
     *
     * @param wordSourceId the ID of the document index whose collection reference count is retrieved.
     *
     * @return the reference count of the document index.
     *
     * @throws IndexPersistenceException if retrieving the collection reference count fails; or the document index is
     *         not persisted.
     */
    int getDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        // get the document id from the word source id
        String documentId = Util.getDocumentId(wordSourceId);

        Connection connection = persistence.getConnection();
        PreparedStatement pstmt = null;
        ResultSet rs = null;

        try {
            // get the use count from the DOCUMENT table
            pstmt = connection.prepareStatement(SQL_SEL_USE_COUNT);
            pstmt.setString(1, documentId);
            rs = pstmt.executeQuery();

            // check if the document index exists
            if (!rs.next()) {
                throw new IndexPersistenceException("Document index does not exist");
            }
            return rs.getInt(1);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to get use count", e);
        } finally {
            DBUtil.close(rs);
            DBUtil.close(pstmt);
            DBUtil.close(connection);
        }
    }

    /**
     * <p>
     * Parses the specified string into a locale. The string should be read from persistence, and have these
     * variations: Language, Language_Country, Language_Country_Version.
     * </p>
     *
     * @param str the string to parse.
     *
     * @return the parsed locale.
     *
     * @throws IndexPersistenceException if the string is not a valid locale string.
     */
    private static Locale parseLocale(String str) throws IndexPersistenceException {
        String[] parts = str.split("_");

        if (parts.length == 1) {
            // the format is Language
            return new Locale(parts[0]);
        } else if (parts.length == 2) {
            // the format is Language_Country
            return new Locale(parts[0], parts[1]);
        } else if (parts.length == 3) {
            // the format is Language_Country_Version
            return new Locale(parts[0], parts[1], parts[2]);
        } else {
            // the format is not recognized
            throw new IndexPersistenceException("Locale string " + str + " is not valid");
        }
    }
}
