/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.db.connectionfactory.DBConnectionFactory;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.persistence.impl.Util;
import com.topcoder.document.index.wordsource.WordSourceId;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import java.text.CollationKey;
import java.text.Collator;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * <p>
 * This is an implementation of the AbstractDBIndexPersistence that deals with a fast, CLOB based, normalized database
 * schema. This is provided for a user who is concerned about speed of the execution. By storing all the word and word
 * position data in a single CLOB structure we are achieving a 4000% speed increase over the simple access solution.
 * However the data mining capabilities are limited.
 * </p>
 *
 * <p>
 * This class is not thread-safe. The methods should not be called concurrently on a single instance. However, the
 * database operations (add, remove and update) are made atomic using JDBC transactions. So multiple instances of this
 * class can access the database at the same time.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
public class FastAccessDBIndexPersistence extends AbstractDBIndexPersistence {
    /**
     * <p>
     * The SQL statement to insert the word data.
     * </p>
     */
    private static final String SQL_INS_WORD_DATA = "INSERT INTO WORD_DATA VALUES (?, ?)";

    /**
     * <p>
     * The SQL statement to delete the word data belonging to a document.
     * </p>
     */
    private static final String SQL_DEL_WORD_DATA = "DELETE FROM WORD_DATA WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * The SQL statement to select the word data belonging to a document.
     * </p>
     */
    private static final String SQL_SEL_WORD_DATA = "SELECT WORD_DATA FROM WORD_DATA WHERE DOCUMENT_ID = ?";

    /**
     * <p>
     * The underlying persistence providing some common operations related to database persistence of document index.
     * This cannot be changed and cannot be null.
     * </p>
     */
    private final DocumentIndexPersistence documentIndexPersistence = new DocumentIndexPersistence(this);

    /**
     * <p>
     * The underlying persistence providing some common operations related to database persistence of collection index.
     * This cannot be changed and cannot be null.
     * </p>
     */
    private final CollectionIndexPersistence collectionIndexPersistence = new CollectionIndexPersistence(this);

    /**
     * <p>
     * A constructor that loads the connection-related properties from the specified namespace. Simply calls the super
     * constructor.
     * </p>
     *
     * @param namespace The namespace for connection-related properties.
     *
     * @throws IllegalArgumentException if namespace is null or empty string.
     * @throws PersistenceConfigurationException if there are configuration issues encountered.
     */
    public FastAccessDBIndexPersistence(String namespace)
        throws PersistenceConfigurationException {
        super(namespace);
    }

    /**
     * <p>
     * A constructor that accepts the given connection factory and connection name. Simply calls the super constructor.
     * If connectionName is null or empty string, the default connection will be used.
     * </p>
     *
     * @param connectionFactory the connection factory used to generate connections.
     * @param connectionName the name of the connection used for persistence operations.
     *
     * @throws IllegalArgumentException if connectionFactory is null.
     */
    public FastAccessDBIndexPersistence(DBConnectionFactory connectionFactory, String connectionName) {
        super(connectionFactory, connectionName);
    }

    /**
     * <p>
     * Adds the given document index to the persistence. The initial collection reference count should be 0. When the
     * document index exists in the persistence, it fails.
     * </p>
     *
     * <p>
     * This method adds records to the DOCUMENT, DELIMITER and WORD_DATA tables.
     * </p>
     *
     * @param documentIndex the document index to be persisted.
     *
     * @throws IllegalArgumentException if documentIndex is null.
     * @throws IndexPersistenceException if persisting the document index fails; or the document index exists in the
     *         persistence.
     */
    public void addDocumentIndex(DocumentIndex documentIndex)
        throws IndexPersistenceException {
        Util.checkDocumentIndex(documentIndex);

        // get the document id from the word source id
        WordSourceId wordSourceId = documentIndex.getWordSourceId();
        String documentId = Util.getDocumentId(wordSourceId);

        Connection connection = getConnection();
        PreparedStatement pstmt = null;
        Reader reader = null;

        try {
            // start the transaction
            beginTransaction(connection);

            // add records to the DOCUMENT and DELIMITER tables
            documentIndexPersistence.addDocumentIndex(connection, documentId, documentIndex);

            // perform the specific handling for words and word positions
            Map words = documentIndex.getWords();
            WordDataWriter wordDataWriter = new WordDataWriter(wordSourceId.getDelimiters());

            for (Iterator i = words.entrySet().iterator(); i.hasNext();) {
                // extract each word and its positions
                Map.Entry entry = (Map.Entry) i.next();
                String word = ((CollationKey) entry.getKey()).getSourceString();
                List positions = (List) entry.getValue();

                // write the word and its positions to form CLOB word data
                wordDataWriter.writeWord(word);
                wordDataWriter.writeWordPositions(positions);
            }
            reader = new StringReader(wordDataWriter.getWordData());

            // add the word data to the WORD_DATA table
            // we use setCharacterStream() for word data since it will be long, and it supports CLOB as well:
            // http://publib.boulder.ibm.com/infocenter/idshelp/v10/topic/com.ibm.jdbc.doc/jdbc218.htm
            pstmt = connection.prepareStatement(SQL_INS_WORD_DATA);
            pstmt.setString(1, documentId);
            pstmt.setCharacterStream(2, reader, wordDataWriter.getWordDataLength());
            pstmt.executeUpdate();
            pstmt.close();

            // commit the transaction
            commitTransaction(connection);
        } catch (IndexPersistenceException e) {
            rollbackTransaction(connection);
            throw e;
        } catch (SQLException e) {
            rollbackTransaction(connection);
            throw new IndexPersistenceException("Fails to add document index", e);
        } finally {
            Util.close(reader);
        }
    }

    /**
     * <p>
     * Gets the document index from the persistence. The ID of the document is given. When the document is not
     * persisted in the persistence, null should be returned.
     * </p>
     *
     * <p>
     * This method retrieves records from the WORD_DATA table. It does not deal with the DOCUMENT and DELIMITER tables
     * since the given wordSourceId already provides the information.
     * </p>
     *
     * @param wordSourceId the ID of the document whose index is retrieved.
     *
     * @return the document index of the document with the ID, or null if not found.
     *
     * @throws IllegalArgumentException if wordSourceId is null.
     * @throws IndexPersistenceException if retrieving the document index fails.
     */
    public DocumentIndex getDocumentIndex(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        Util.checkWordSourceId(wordSourceId);

        // get the document id from the word source id
        String documentId = Util.getDocumentId(wordSourceId);

        // prepare the collator under the current locale
        Collator collator = Collator.getInstance(wordSourceId.getSourceLocale());

        Connection connection = getConnection();
        PreparedStatement pstmt = null;
        ResultSet rs = null;
        Reader reader = null;

        try {
            // check if the document index exists
            if (!documentIndexPersistence.existDocumentIndex(connection, documentId)) {
                return null;
            }

            // get the word data from the WORD_DATA table
            // we use getCharacterStream() for word data since it will be long, and it supports CLOB as well:
            // http://publib.boulder.ibm.com/infocenter/idshelp/v10/topic/com.ibm.jdbc.doc/jdbc218.htm
            pstmt = connection.prepareStatement(SQL_SEL_WORD_DATA);
            pstmt.setString(1, documentId);
            rs = pstmt.executeQuery();

            // prepare the mapping from collation keys to word positions
            Map words = new HashMap();

            if (rs.next()) {
                reader = new BufferedReader(rs.getClob(1).getCharacterStream());

                // read the words and word positions from the CLOB word data
                WordDataReader wordDataReader = new WordDataReader(reader, wordSourceId.getDelimiters());
                String word;

                while ((word = wordDataReader.readWord()) != null) {
                    // add the list of word positions to the map
                    words.put(collator.getCollationKey(word), wordDataReader.readWordPositions());
                }
            }
            return new DocumentIndex(wordSourceId, words);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to get document index", e);
        } finally {
            Util.close(reader);
            DBUtil.close(rs);
            DBUtil.close(pstmt);
            DBUtil.close(connection);
        }
    }

    /**
     * <p>
     * Removes a document index from the persistence. The document index can be deleted only when there is no more
     * collection containing the document index. The ID of the document is given. When the document is not persisted,
     * it fails.
     * </p>
     *
     * <p>
     * This method deletes records from the DOCUMENT, DELIMITER and WORD_DATA tables.
     * </p>
     *
     * @param wordSourceId the ID of the document whose index is deleted.
     *
     * @throws IllegalArgumentException if wordSourceId is null.
     * @throws IndexPersistenceException if deleting the document index fails; or the document index is not persisted;
     *         or there is at least one collection containing the document index.
     */
    public void removeDocumentIndex(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        Util.checkWordSourceId(wordSourceId);

        // get the document id from the word source id
        String documentId = Util.getDocumentId(wordSourceId);

        Connection connection = getConnection();

        try {
            // start the transaction
            beginTransaction(connection);

            // check if the documnt index exists and is unused
            if (!documentIndexPersistence.existUnusedDocumentIndex(connection, documentId)) {
                throw new IndexPersistenceException("Unused document index does not exist");
            }

            // remove the word data from the WORD_DATA table
            DBUtil.execute(connection, SQL_DEL_WORD_DATA, documentId);

            // remove records from the DOCUMENT and DELIMITER tables
            documentIndexPersistence.removeDocumentIndex(connection, documentId);

            // commit the transaction
            commitTransaction(connection);
        } catch (IndexPersistenceException e) {
            rollbackTransaction(connection);
            throw e;
        } catch (SQLException e) {
            rollbackTransaction(connection);
            throw new IndexPersistenceException("Fails to remove document index", e);
        }
    }

    /**
     * <p>
     * Adds the collection index to the persistence. When the collection index exists in the persistence, it fails.
     * When the document indexes of the collection index are not persisted, it also fails.
     * </p>
     *
     * <p>
     * This method adds records to the DOCUMENT_COLLECTION and DOCUMENT_COLLECTION_DOCUMENT_XREF tables. The use count
     * is not updated here, since it is handled by the CollectionIndex class.
     * </p>
     *
     * @param collectionIndex the collection index to be persisted.
     *
     * @throws IllegalArgumentException if collectionIndex is null.
     * @throws IndexPersistenceException if persisting the collection index fails; or the collection index exists in
     *         the persistence; or the document indexes of the collection index are not persisted.
     */
    public void addCollectionIndex(CollectionIndex collectionIndex)
        throws IndexPersistenceException {
        Util.checkCollectionIndex(collectionIndex);
        collectionIndexPersistence.addCollectionIndex(collectionIndex);
    }

    /**
     * <p>
     * Gets the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, null should be returned.
     * </p>
     *
     * <p>
     * This method retrieves records from the DOCUMENT_COLLECTION, DOCUMENT_COLLECTION_DOCUMENT_XREF, DOCUMENT,
     * DELIMITER and WORD_DATA tables. Note that the collection index is created with null owner. It should be set by
     * the IndexManager using the setOwner() method.
     * </p>
     *
     * @param collectionId the ID of the collection index to be retrieved.
     *
     * @return the collection index with the given ID.
     *
     * @throws IllegalArgumentException if collectionId is null or empty string.
     * @throws IndexPersistenceException if retrieving the collection index fails.
     */
    public CollectionIndex getCollectionIndex(String collectionId)
        throws IndexPersistenceException {
        Util.checkCollectionId(collectionId);

        Connection connection = getConnection();
        PreparedStatement pstmt = null;
        ResultSet rs = null;
        Reader reader = null;

        try {
            // check if the collection index exists
            if (!collectionIndexPersistence.existCollectionIndex(connection, collectionId)) {
                return null;
            }

            // get the document IDs in the collection
            Set wordSourceIds = collectionIndexPersistence.getWordSourceIds(connection, collectionId);
            Map words = new HashMap();

            // perform the specific handling for words
            pstmt = connection.prepareStatement(SQL_SEL_WORD_DATA);

            for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
                // extract each word source id and its collator
                WordSourceId wordSourceId = (WordSourceId) i.next();
                String documentId = Util.getDocumentId(wordSourceId);
                Collator collator = Collator.getInstance(wordSourceId.getSourceLocale());

                // get the words from the WORD_DATA table
                // then free up the resources immediately since the number of words is large
                pstmt.setString(1, documentId);
                rs = pstmt.executeQuery();

                if (rs.next()) {
                    reader = new BufferedReader(rs.getClob(1).getCharacterStream());

                    // read the words from the CLOB word data
                    WordDataReader wordDataReader = new WordDataReader(reader, wordSourceId.getDelimiters());
                    String word;

                    while ((word = wordDataReader.readWord()) != null) {
                        CollationKey key = collator.getCollationKey(word);
                        Set value = (Set) words.get(key);

                        // if the word has not been added, add a new set to the map
                        if (value == null) {
                            value = new HashSet();
                            words.put(key, value);
                        }

                        // add the word source id to the set
                        value.add(wordSourceId);
                    }
                    reader.close();
                }
                rs.close();
            }

            // create the collection index with null owner
            // it should be set by the IndexManager using the setOwner() method
            return new CollectionIndex(null, words, wordSourceIds, collectionId);
        } catch (SQLException e) {
            throw new IndexPersistenceException("Fails to get collection index", e);
        } catch (IOException e) {
            throw new IndexPersistenceException("Some I/O error occurs", e);
        } finally {
            Util.close(reader);
            DBUtil.close(rs);
            DBUtil.close(pstmt);
            DBUtil.close(connection);
        }
    }

    /**
     * <p>
     * Removes the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, it fails.
     * </p>
     *
     * <p>
     * This method deletes records from the DOCUMENT_COLLECTION and DOCUMENT_COLLECTION_DOCUMENT_XREF tables. The use
     * count is not updated here, since it is handled by the CollectionIndex class.
     * </p>
     *
     * @param collectionId the ID of the collection index to be deleted.
     *
     * @throws IllegalArgumentException if collectionId is null or empty string.
     * @throws IndexPersistenceException if deleting the collection index fails; or the collection index ID cannot be
     *         found in the persistence.
     */
    public void removeCollectionIndex(String collectionId)
        throws IndexPersistenceException {
        Util.checkCollectionId(collectionId);
        collectionIndexPersistence.removeCollectionIndex(collectionId);
    }

    /**
     * <p>
     * Updates the collection index in the persistence. When the collection index ID cannot be found in the
     * persistence, it fails.
     * </p>
     *
     * <p>
     * This method updates records in the DOCUMENT_COLLECTION_DOCUMENT_XREF table. The use count is not updated here,
     * since it is handled by the CollectionIndex class.
     * </p>
     *
     * @param collectionIndex the collection index to be updated.
     *
     * @throws IllegalArgumentException if collectionIndex is null.
     * @throws IndexPersistenceException if updating the collection index fails; or the collection index ID cannot be
     *         found in the persistence.
     */
    public void updateCollectionIndex(CollectionIndex collectionIndex)
        throws IndexPersistenceException {
        Util.checkCollectionIndex(collectionIndex);
        collectionIndexPersistence.updateCollectionIndex(collectionIndex);
    }

    /**
     * <p>
     * Increases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * <p>
     * This method updates a record in the DOCUMENT table.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is increased by 1.
     *
     * @throws IllegalArgumentException if wordSourceId is null.
     * @throws IndexPersistenceException if increasing the collection reference count fails; or the document index is
     *         not persisted.
     */
    public void increaseDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        Util.checkWordSourceId(wordSourceId);
        documentIndexPersistence.increaseDocumentUseCount(wordSourceId);
    }

    /**
     * <p>
     * Decreases the collection reference count of the document index. The document ID is given. When the document
     * index is not persisted, it fails.
     * </p>
     *
     * <p>
     * This method updates a record in the DOCUMENT table.
     * </p>
     *
     * @param wordSourceId the ID of the document whose collection reference count is decreased by 1.
     *
     * @throws IllegalArgumentException if wordSourceId is null.
     * @throws IndexPersistenceException if decreasing the collection reference count fails; or the document index is
     *         not persisted.
     */
    public void decreaseDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        Util.checkWordSourceId(wordSourceId);
        documentIndexPersistence.decreaseDocumentUseCount(wordSourceId);
    }

    /**
     * <p>
     * Gets a set of all IDs of documents whose indexes are persisted in the persistence.
     * </p>
     *
     * <p>
     * This method retrieves records from the DOCUMENT and DELIMITER tables.
     * </p>
     *
     * @return a set of all IDs of persisted document indexes.
     *
     * @throws IndexPersistenceException if retrieving the document IDs fails.
     */
    public Set getIndexedDocuments() throws IndexPersistenceException {
        return documentIndexPersistence.getIndexedDocuments();
    }

    /**
     * <p>
     * Gets the collection reference count of the given document index. When the document index is not persisted, it
     * fails.
     * </p>
     *
     * <p>
     * This method retrieves a record from the DOCUMENT table.
     * </p>
     *
     * @param wordSourceId the ID of the document index whose collection reference count is retrieved.
     *
     * @return the reference count of the document index.
     *
     * @throws IllegalArgumentException if wordSourceId is null.
     * @throws IndexPersistenceException if retrieving the collection reference count fails; or the document index is
     *         not persisted.
     */
    public int getDocumentUseCount(WordSourceId wordSourceId)
        throws IndexPersistenceException {
        Util.checkWordSourceId(wordSourceId);
        return documentIndexPersistence.getDocumentUseCount(wordSourceId);
    }

    /**
     * Private inner class to parse the words and word positions from the CLOB word data. Basic operation is very
     * simple. The readWord() method is called to get the next word available. If desired the corresponding word
     * positions can be obtained by calling the readWordPositions() method. This class is inner since it is used by
     * fast-access database only. The supplied arguments are assumed to be valid.
     *
     * @author AleaActaEst, TCSDEVELOPER
     * @version 1.0
     */
    private static class WordDataReader {
        /**
         * The delimiter in the word data structure. Single occurrence represents position delimiter. Double occurrence
         * represents word delimiter.
         */
        private final char delimiter;

        /**
         * The reader to parse the word data from.
         */
        private final Reader reader;

        /**
         * The word positions of the current word. It is a list of integers representing the positions.
         */
        private final List wordPositions = new ArrayList();

        /**
         * Constructor that accepts the reader and delimiters. Only the first delimiter in the array is used in the
         * word data.
         *
         * @param reader the reader to parse the word data from.
         * @param delimiters the delimiters used to separate the words.
         */
        public WordDataReader(Reader reader, String[] delimiters) {
            this.reader = reader;
            delimiter = delimiters[0].charAt(0);
        }

        /**
         * Returns the word positions of the current word.
         *
         * @return a list of integers representing the positions.
         */
        public List readWordPositions() {
            return new ArrayList(wordPositions);
        }

        /**
         * Returns the next word available from the reader. The corresponding word positions are updated after calling
         * this method. If no more word is available, null is returned.
         *
         * @return the next word, or null if no word is available.
         *
         * @throws IndexPersistenceException if any I/O error occurs, or the word data is corrupted.
         */
        public String readWord() throws IndexPersistenceException {
            // initially the reader starts with word
            StringBuffer word = new StringBuffer();
            StringBuffer wordPosition = new StringBuffer();
            StringBuffer current = word;

            // clear the previous word positions
            wordPositions.clear();

            try {
                char prev = 0;
                char next = 0;
                int read;

                while ((read = reader.read()) != -1) {
                    next = (char) read;
                    if (next == delimiter) {
                        if (prev == delimiter) {
                            // if word delimiter is encountered, stop further parsing
                            break;
                        } else if (current == wordPosition) {
                            // if position delimiter is encountered, convert the previous word position to number
                            wordPositions.add(new Integer(wordPosition.toString()));
                            wordPosition.setLength(0);
                        } else {
                            // otherwise switch the parsing from word to position
                            current = wordPosition;
                        }
                    } else {
                        current.append(next);
                    }
                    prev = next;
                }

                // convert the last word position to number
                if ((read == -1) && (wordPosition.length() > 0)) {
                    wordPositions.add(new Integer(wordPosition.toString()));
                }
            } catch (IOException e) {
                throw new IndexPersistenceException("Fails to read next word");
            } catch (NumberFormatException e) {
                throw new IndexPersistenceException("Word position is not a number");
            }

            // check if the word or word position data is corrupted
            if ((word.length() == 0) && !wordPositions.isEmpty()) {
                throw new IndexPersistenceException("Word is missing");
            }
            if ((word.length() > 0) && wordPositions.isEmpty()) {
                throw new IndexPersistenceException("Word position is missing");
            }

            return (word.length() == 0) ? null : word.toString();
        }
    }

    /**
     * Private inner class to write the words and word positions to the CLOB word data. Basic operation is very simple.
     * The writeWord() is called to write the next word. The writeWordPositions() is then called to write the word
     * positions of that word. This process repeats for each word. This class is inner since it is used by fast-access
     * database only. The supplied arguments are assumed to be valid.
     *
     * @author AleaActaEst, TCSDEVELOPER
     * @version 1.0
     */
    private static class WordDataWriter {
        /**
         * The delimiter in the word data structure. Single occurrence represents position delimiter. Double occurrence
         * represents word delimiter.
         */
        private final char delimiter;

        /**
         * The buffer to store the word data output.
         */
        private final StringBuffer buffer = new StringBuffer();

        /**
         * Constructor that accepts the delimiters. Only the first delimiter in the array is used in the word data.
         *
         * @param delimiters the delimiters used to separate the words.
         */
        public WordDataWriter(String[] delimiters) {
            delimiter = delimiters[0].charAt(0);
        }

        /**
         * Writes the word positions of the current word to the word data.
         *
         * @param wordPositions a list of integers representing the positions.
         */
        public void writeWordPositions(List wordPositions) {
            for (Iterator i = wordPositions.iterator(); i.hasNext();) {
                // use single delimiter
                buffer.append(delimiter);
                buffer.append(((Integer) i.next()).intValue());
            }
        }

        /**
         * Writes the next word to the word data.
         *
         * @param word the next word.
         */
        public void writeWord(String word) {
            // use double delimiters
            if (buffer.length() > 0) {
                buffer.append(delimiter);
                buffer.append(delimiter);
            }
            buffer.append(word);
        }

        /**
         * Returns the resulting word data in a string.
         *
         * @return the word data.
         */
        public String getWordData() {
            return buffer.toString();
        }

        /**
         * Returns the number of characters in the resulting word data.
         *
         * @return the number of characters in the resulting word data.
         */
        public int getWordDataLength() {
            return buffer.length();
        }
    }
}
