/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.xml;

import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.Util;
import com.topcoder.document.index.wordsource.WordSourceId;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import java.text.CollationKey;
import java.text.Collator;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;


/**
 * <p>
 * A parser to extract information in the document index file using SAX. The extracted information includes the ID of
 * the document (WordSourceId instance), and the mapping from words to their positions. If the document index has
 * several sequence files, each has to be processed by this parser.
 * </p>
 *
 * <p>
 * Note that the parser does not verify the structure of the xml file. The contents of the files are assumed to be
 * maintained by the XmlIndexPersistence class, so consistency is assured.
 * </p>
 *
 * <p>
 * This class is package-private, so that they can only be used by classes in the same package.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
class DocumentIndexParser extends DefaultHandler {
    /**
     * <p>
     * A flag indicating the next character data has nothing of interest.
     * </p>
     */
    private static final int PARSE_NOTHING = 1;

    /**
     * <p>
     * A flag indicating the next character data represents the source identity.
     * </p>
     */
    private static final int PARSE_SOURCE_IDENTITY = 2;

    /**
     * <p>
     * A flag indicating the next character data represents the class name.
     * </p>
     */
    private static final int PARSE_CLASS_NAME = 3;

    /**
     * <p>
     * A flag indicating the next character data represents the word.
     * </p>
     */
    private static final int PARSE_WORD = 4;

    /**
     * <p>
     * A flag indicating the next character data represents the word position.
     * </p>
     */
    private static final int PARSE_WORD_POSITION = 5;

    /**
     * <p>
     * The source identity of the document index. It is initially set to null. It can be changed when the xml file is
     * parsed.
     * </p>
     */
    private Object sourceIdentity = null;

    /**
     * <p>
     * The class name of the document source used to read words of the document. It is initially set to null. It can be
     * changed when the xml file is parsed.
     * </p>
     */
    private String className = null;

    /**
     * <p>
     * The locale of the document. It is initially set to null. It can be changed when the xml file is parsed.
     * </p>
     */
    private Locale locale = null;

    /**
     * <p>
     * The delimiters (String) used to separate the words. It cannot be changed. The delimiters are populated when the
     * xml file is parsed.
     * </p>
     */
    private final List delimiters = new ArrayList();

    /**
     * <p>
     * The word map of the indexed data. The keys of the word map are collation keys of all different words. The values
     * are lists of integers representing the positions of the words in the document. It cannot be changed. The words
     * are populated when the xml file is parsed.
     * </p>
     */
    private final Map words = new HashMap();

    /**
     * <p>
     * A flag indicating what the next character data represents. It can be changed when the xml file is parsed.
     * </p>
     */
    private int parseNext = PARSE_NOTHING;

    /**
     * <p>
     * The collator to generate the collation keys of all different words. It can be changed once the locale is parsed.
     * </p>
     */
    private Collator collator = null;

    /**
     * <p>
     * The collation key for the last word read. It can be changed when the word tag is parsed.
     * </p>
     */
    private CollationKey lastKey = null;

    /**
     * <p>
     * Creates a new instance of DocumentIndexParser
     * </p>
     */
    public DocumentIndexParser() {
    }

    /**
     * <p>
     * Parses the document index information. This includes the source identity, class name, locale, delimiters, words
     * and word positions. If any of these fields is not available or incorrect, a SAXException will be thrown.
     * </p>
     *
     * @param uri the uri namespace.
     * @param localName the local name (without prefix).
     * @param qName the qualified name (with prefix).
     * @param attributes the specified attributes.
     *
     * @throws SAXException if any parsing error occurs.
     */
    public void startElement(String uri, String localName, String qName, Attributes attributes)
        throws SAXException {
        if (qName.equals(XmlConstants.SOURCE_IDENTITY_TAG_NAME)) {
            // character data points to source identity
            parseNext = PARSE_SOURCE_IDENTITY;
        } else if (qName.equals(XmlConstants.CLASS_NAME_TAG_NAME)) {
            // character data points to class name
            parseNext = PARSE_CLASS_NAME;
        } else if (qName.equals(XmlConstants.WORD_TAG_NAME)) {
            // character data points to word
            parseNext = PARSE_WORD;
        } else if (qName.equals(XmlConstants.POS_TAG_NAME)) {
            // character data points to word position
            parseNext = PARSE_WORD_POSITION;
        } else if (qName.equals(XmlConstants.LOCALE_TAG_NAME)) {
            // parse the locale of the document
            String language = attributes.getValue(XmlConstants.LANGUAGE_ATTR_NAME);
            String country = attributes.getValue(XmlConstants.COUNTRY_ATTR_NAME);
            String variant = attributes.getValue(XmlConstants.VARIANT_ATTR_NAME);

            if ((language != null) && (country == null) && (variant == null)) {
                // only the language is specified
                locale = new Locale(language);
            } else if ((language != null) && (country != null) && (variant == null)) {
                // the language and country are specified
                locale = new Locale(language, country);
            } else if ((language != null) && (country != null) && (variant != null)) {
                // the language, country and variant are specified
                locale = new Locale(language, country, variant);
            } else {
                // otherwise, the specification is not recognized
                throw new SAXException(XmlConstants.LOCALE_TAG_NAME + " has invalid attributes");
            }
            collator = Collator.getInstance(locale);
        } else if (qName.equals(XmlConstants.ENTRY_TAG_NAME)) {
            // parse the delimiter
            String delimiter = attributes.getValue(XmlConstants.DELIMITER_ATTR_NAME);

            // check if any of the fields is missing
            if (delimiter == null) {
                throw new SAXException(XmlConstants.DELIMITER_ATTR_NAME + " is missing");
            }

            // add the delimiter to the list
            delimiters.add(delimiter);
        }
    }

    /**
     * <p>
     * Determines when the tag ends for the particular field.
     * </p>
     *
     * @param uri the uri namespace.
     * @param localName the local name (without prefix).
     * @param qName the qualified name (with prefix).
     */
    public void endElement(String uri, String localName, String qName) {
        parseNext = PARSE_NOTHING;
    }

    /**
     * <p>
     * Parses the document index information. This includes the source identity, class name, words and word positions.
     * If any of these fields is not available or incorrect, a SAXException will be thrown.
     * </p>
     *
     * @param ch the character array.
     * @param start the start position in the character array.
     * @param length the number of characters to use from the character array.
     *
     * @throws SAXException if any parsing error occurs.
     */
    public void characters(char[] ch, int start, int length)
        throws SAXException {
        String str = new String(ch, start, length);

        if (parseNext == PARSE_SOURCE_IDENTITY) {
            // parse the source identity - decompress it from Base64 representation
            try {
                sourceIdentity = Util.decompress(str);
            } catch (IndexPersistenceException e) {
                throw new SAXException("Fails to decompress source identity");
            }
        } else if (parseNext == PARSE_CLASS_NAME) {
            // parse the class name
            className = str;
        } else if (parseNext == PARSE_WORD) {
            // parse the word - locale must be present
            if (collator == null) {
                throw new SAXException("Missing " + XmlConstants.LOCALE_TAG_NAME + " for "
                    + XmlConstants.WORD_TAG_NAME);
            }
            lastKey = collator.getCollationKey(str);
        } else if (parseNext == PARSE_WORD_POSITION) {
            // parse the word position - word must be present
            if (lastKey == null) {
                throw new SAXException("Missing " + XmlConstants.WORD_TAG_NAME + " for " + XmlConstants.POS_TAG_NAME);
            }

            List positions = (List) words.get(lastKey);

            // if the word has not been added, add a new list to the map
            if (positions == null) {
                positions = new ArrayList();
                words.put(lastKey, positions);
            }

            // add the word position to the list
            try {
                positions.add(new Integer(str));
            } catch (NumberFormatException e) {
                throw new SAXException(XmlConstants.POS_TAG_NAME + " is not a number");
            }
        }
    }

    /**
     * <p>
     * Returns the ID of the document index persisted in the document index file.
     * </p>
     *
     * <p>
     * This method should be called after the xml file is parsed.
     * </p>
     *
     * @return the ID of the document index.
     */
    public WordSourceId getWordSourceId() {
        return new WordSourceId(sourceIdentity, className, (String[]) delimiters.toArray(new String[0]), locale);
    }

    /**
     * <p>
     * Returns the word map of the indexed data. The keys of the word map are collation keys of all different words.
     * The values are lists of integers representing the positions of the words in the document. May need to
     * concatenate the word maps if the document index spans across several sequence files. A copy of the inner map is
     * returned.
     * </p>
     *
     * <p>
     * This method should be called after the xml file is parsed.
     * </p>
     *
     * @return the word map of the indexed data.
     */
    public Map getWords() {
        return new HashMap(words);
    }
}
