/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.xml;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;


/**
 * <p>
 * A parser to extract information in the document usage file using SAX. The extracted information includes the
 * document id and its use count. The existence of a document index can be quickly indicated by its record in the
 * document usage file.
 * </p>
 *
 * <p>
 * Note that the parser does not verify the consistency of the document indexes with the collection indexes, as well as
 * the structure of the xml file. The contents of the files are assumed to be maintained by the XmlIndexPersistence
 * class, so consistency is assured.
 * </p>
 *
 * <p>
 * This class is package-private, so that they can only be used by classes in the same package.
 * </p>
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
class DocumentUsageParser extends DefaultHandler {
    /**
     * <p>
     * A mapping of document ids (String) to its usage count (Integer). This cannot be changed. The mappings are
     * populated once the xml file is parsed.
     * </p>
     */
    private final Map documentIndexes = new HashMap();

    /**
     * <p>
     * Creates a new instance of DocumentUsageParser.
     * </p>
     */
    public DocumentUsageParser() {
    }

    /**
     * <p>
     * Parses the document id and usage count information. If any of these fields is not available or incorrect, a
     * SAXException will be thrown.
     * </p>
     *
     * @param uri the uri namespace.
     * @param localName the local name (without prefix).
     * @param qName the qualified name (with prefix).
     * @param attributes the specified attributes.
     *
     * @throws SAXException if any parsing error occurs.
     */
    public void startElement(String uri, String localName, String qName, Attributes attributes)
        throws SAXException {
        if (qName.equals(XmlConstants.DOCUMENT_USAGE_TAG_NAME)) {
            // extract the document id and usage count
            String documentId = attributes.getValue(XmlConstants.DOCUMENT_ID_ATTR_NAME);
            String usageCount = attributes.getValue(XmlConstants.USAGE_COUNT_ATTR_NAME);

            // check if any of the fields is missing
            if (documentId == null) {
                throw new SAXException(XmlConstants.DOCUMENT_ID_ATTR_NAME + " is missing");
            }
            if (usageCount == null) {
                throw new SAXException(XmlConstants.USAGE_COUNT_ATTR_NAME + " is missing");
            }

            try {
                // add the document to the map
                documentIndexes.put(documentId, new Integer(usageCount));
            } catch (NumberFormatException e) {
                throw new SAXException(XmlConstants.USAGE_COUNT_ATTR_NAME + " is not a number");
            }
        }
    }

    /**
     * <p>
     * Returns the set of all document ids in the document usage file. A copy of the inner set is returned.
     * </p>
     *
     * <p>
     * This method should be called after the xml file is parsed.
     * </p>
     *
     * @return the set of all document ids.
     */
    public Set getAllDocumentIds() {
        return new HashSet(documentIndexes.keySet());
    }

    /**
     * <p>
     * Returns the usage count of the document index. The unique document id is given. If the document index is not
     * persisted, -1 is returned.
     * </p>
     *
     * <p>
     * This method should be called after the xml file is parsed.
     * </p>
     *
     * @param documentId the unique document id whose index is to be checked.
     *
     * @return the usage count of the document index, or -1 if the document index is not persisted.
     */
    public int getUsageCount(String documentId) {
        Integer usageCount = (Integer) documentIndexes.get(documentId);

        return (usageCount == null) ? (-1) : usageCount.intValue();
    }
}
