/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.wordsource.WordSourceId;

import java.text.Collator;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;


/**
 * A helper class to provide some useful methods for testing. These include creating various types of document index
 * and collection index.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class TestHelper {
    /**
     * Private constructor to prevent instantiation.
     */
    private TestHelper() {
    }

    /**
     * Creates an ID of the document index with the specified source identity for testing. Other fields are fixed.
     *
     * @param sourceIdentity the identity of the document.
     *
     * @return an ID of the document index.
     */
    public static WordSourceId createWordSourceId(Object sourceIdentity) {
        String className = "className";
        String[] delimiters = {"\t", "\n", "\r"};
        Locale locale = new Locale("en");

        return new WordSourceId(sourceIdentity, className, delimiters, locale);
    }

    /**
     * Creates an illegal ID of the document index for testing. The source identity is non-serializable.
     *
     * @return an illegal ID of the document index.
     */
    public static WordSourceId createIllegalWordSourceId() {
        return createWordSourceId(new Object());
    }

    /**
     * Creates a valid ID of the document index for testing. The source identity is set to a string.
     *
     * @return a valid ID of the document index.
     */
    public static WordSourceId createWordSourceId() {
        return createWordSourceId("sourceIdentity");
    }

    /**
     * Creates the words map for the document index for testing. Three words each with three positions are added to the
     * words map.
     *
     * @return the words map for the document index.
     */
    public static Map createDocumentIndexWords() {
        Collator collator = Collator.getInstance(Locale.US);
        String[] words = {"document", "index", "persistence"};

        Map allWords = new LinkedHashMap();

        // prepare the words mapping
        for (int i = 0; i < words.length; i++) {
            int position = (i * 3) + 1;

            allWords.put(collator.getCollationKey(words[i]), toList(new int[] {position, position + 1, position + 2}));
        }
        return allWords;
    }

    /**
     * Creates the words map for the long document index for testing. Ten words each with three positions are added to
     * the words map.
     *
     * @return the words map for the document index.
     */
    public static Map createLongDocumentIndexWords() {
        Collator collator = Collator.getInstance(Locale.US);

        Map allWords = new LinkedHashMap();
        int position = 1;

        // prepare the words mapping
        for (int i = 0; i < 10; i++) {
            allWords.put(collator.getCollationKey("document" + i),
                toList(new int[] {position++, position++, position++}));
        }
        return allWords;
    }

    /**
     * Creates a long document index for testing. Ten words each with three positions are added to the document index.
     *
     * @return a long document index.
     */
    public static DocumentIndex createLongDocumentIndex() {
        return new DocumentIndex(createWordSourceId(), createLongDocumentIndexWords());
    }

    /**
     * Creates a document index with the specified source identity for testing. Three words each with three positions
     * are added to the document index.
     *
     * @param sourceIdentity the identity of the document.
     *
     * @return a document index.
     */
    public static DocumentIndex createDocumentIndex(Object sourceIdentity) {
        return new DocumentIndex(createWordSourceId(sourceIdentity), createDocumentIndexWords());
    }

    /**
     * Creates an illegal document index for testing. The source identity is non-serializable.
     *
     * @return an illegal document index.
     */
    public static DocumentIndex createIllegalDocumentIndex() {
        return createDocumentIndex(new Object());
    }

    /**
     * Creates a valid document index for testing. Three words each with three positions are added to the document
     * index.
     *
     * @return a valid document index.
     */
    public static DocumentIndex createDocumentIndex() {
        return createDocumentIndex("sourceIdentity");
    }

    /**
     * Creates the words map for the collection index for testing. Three words each with three document IDs are added
     * to the words map.
     *
     * @return the words map for the collection index.
     */
    public static Map createCollectionIndexWords() {
        Collator collator = Collator.getInstance(Locale.US);
        String[] words = {"document", "index", "persistence"};

        Map allWords = new HashMap();
        Set allDocumentIds = createCollectionIndexDocumentIds();

        for (int i = 0; i < words.length; i++) {
            allWords.put(collator.getCollationKey(words[i]), allDocumentIds);
        }
        return allWords;
    }

    /**
     * Creates the document IDs for the collection index for testing.
     *
     * @return the document IDs for the collection index.
     */
    public static Set createCollectionIndexDocumentIds() {
        Set allDocumentIds = new HashSet();

        for (int i = 0; i < 3; i++) {
            allDocumentIds.add(createWordSourceId("sourceIdentity" + i));
        }
        return allDocumentIds;
    }

    /**
     * Creates a valid collection index for testing. Three IDs of the documents are added to the collection index.
     *
     * @return a valid collection index.
     */
    public static CollectionIndex createCollectionIndex() {
        return new CollectionIndex(null, createCollectionIndexWords(), createCollectionIndexDocumentIds(),
            "collection");
    }

    /**
     * Converts an int array to a list consisting of Integers.
     *
     * @param array an int array.
     *
     * @return an Integer list.
     */
    private static List toList(int[] array) {
        List list = new ArrayList();

        for (int i = 0; i < array.length; i++) {
            list.add(new Integer(array[i]));
        }
        return list;
    }
}
