/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import com.topcoder.document.index.persistence.ConfigHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;

import junit.framework.Test;
import junit.framework.TestSuite;


/**
 * Unit tests for FastAccessDBIndexPersistence implementation.
 *
 * <p>
 * We assume here the instance loaded from configuration namespace will behave the same as created with two-argument
 * constructor. So only the former is used for testing.
 * </p>
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class FastAccessDBIndexPersistenceTest extends BaseDBIndexPersistenceTest {
    /**
     * The sql statement to check a record in the WORD_DATA table.
     */
    private static final String SQL_SEL_WORD_DATA = "SELECT DOCUMENT_ID FROM WORD_DATA WHERE DOCUMENT_ID = ?";

    /**
     * The FastAccessDBIndexPersistence instance to test against.
     */
    private FastAccessDBIndexPersistence persistence = null;

    /**
     * Creates a test suite for the tests in this test case.
     *
     * @return a TestSuite for this test case
     */
    public static Test suite() {
        TestSuite suite = new TestSuite(FastAccessDBIndexPersistenceTest.class);

        return suite;
    }

    /**
     * Prepares a FastAccessDBIndexPersistence instance for testing.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    protected void setUp() throws Exception {
        super.setUp();
        persistence = new FastAccessDBIndexPersistence(ConfigHelper.DB_NAMESPACE);
        setPersistence(persistence);
    }

    /**
     * Clears all the namespaces and database resources.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    protected void tearDown() throws Exception {
        super.tearDown();
    }

    /**
     * Test of constructor with null namespace. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_NullNamespace() throws Exception {
        try {
            new FastAccessDBIndexPersistence(null);
            fail("Creates FastAccessDBIndexPersistence with null namespace");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of constructor with empty namespace. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_EmptyNamespace() throws Exception {
        try {
            new FastAccessDBIndexPersistence("");
            fail("Creates FastAccessDBIndexPersistence with empty namespace");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of constructor with bad namespaces. Expects PersistenceConfigurationException.
     */
    public void testConstructor_BadNamespaces() {
        String[] namespaces = ConfigHelper.BAD_DB_NAMESPACES;

        for (int i = 0; i < namespaces.length; i++) {
            try {
                new FastAccessDBIndexPersistence(namespaces[i]);
                fail("Creates FastAccessDBIndexPersistence with bad namespace");
            } catch (PersistenceConfigurationException e) {
                // good
            }
        }
    }

    /**
     * Test of constructor with null connection factory. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_NullConnectionFactory()
        throws Exception {
        try {
            new FastAccessDBIndexPersistence(null, ConfigHelper.CONNECTION_PRODUCER_NAME);
            fail("Creates FastAccessDBIndexPersistence with null connection factory");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of addDocumentIndex method with non-existing document index. Verifies if the document index was persisted
     * after calling it. This is done by checking the records in the tables.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testAddDocumentIndex_NonExistDocumentIndex()
        throws Exception {
        super.testAddDocumentIndex_NonExistDocumentIndex();

        // check the WORD_DATA table
        parameters.clear();
        parameters.add(documentId);
        assertTrue("Fails to add document index to WORD_DATA table",
            DBHelper.exist(SQL_SEL_WORD_DATA, parameters.toArray()));
    }

    /**
     * Test of removeDocumentIndex method with unused word source id (use count is zero). Verifies if the document
     * index was removed from persistence after calling it. This is done by checking the records in the tables.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testRemoveDocumentIndex_UnusedWordSourceId()
        throws Exception {
        super.testRemoveDocumentIndex_UnusedWordSourceId();

        // check the WORD_DATA table
        assertTrue("Fails to remove document index from WORD_DATA table", DBHelper.isEmptyTable("WORD_DATA"));
    }
}
