/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.xml;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.BaseIndexPersistenceTest;
import com.topcoder.document.index.persistence.ConfigHelper;
import com.topcoder.document.index.persistence.TestHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.wordsource.WordSourceId;

import junit.framework.Test;
import junit.framework.TestSuite;

import java.io.File;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;


/**
 * Unit tests for XmlIndexPersistence implementation.
 *
 * <p>
 * We assume here the instance loaded from configuration namespace will behave the same as created with three-argument
 * constructor. So only the former is used for testing.
 * </p>
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class XmlIndexPersistenceTest extends BaseIndexPersistenceTest {
    /**
     * The XmlIndexPersistence instance to test against.
     */
    private XmlIndexPersistence persistence = null;

    /**
     * Creates a test suite for the tests in this test case.
     *
     * @return a TestSuite for this test case
     */
    public static Test suite() {
        TestSuite suite = new TestSuite(XmlIndexPersistenceTest.class);

        return suite;
    }

    /**
     * Sets up the namespaces and file resources.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    protected void setUp() throws Exception {
        super.setUp();
        ConfigHelper.loadConfig();
        XmlHelper.setupDir();

        persistence = new XmlIndexPersistence(ConfigHelper.XML_NAMESPACE);
        setPersistence(persistence);
    }

    /**
     * Clears all the namespaces and file resources.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    protected void tearDown() throws Exception {
        XmlHelper.cleanupDir();
        ConfigHelper.unloadConfig();
        super.tearDown();
    }

    /**
     * Test of constructor with null namespace. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_NullNamespace() throws Exception {
        try {
            new XmlIndexPersistence(null);
            fail("Creates XmlIndexPersistence with null namespace");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of constructor with empty namespace. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_EmptyNamespace() throws Exception {
        try {
            new XmlIndexPersistence("");
            fail("Creates XmlIndexPersistence with empty namespace");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of constructor with bad namespaces. Expects PersistenceConfigurationException.
     */
    public void testConstructor_BadNamespaces() {
        String[] namespaces = ConfigHelper.BAD_XML_NAMESPACES;

        for (int i = 0; i < namespaces.length; i++) {
            try {
                new XmlIndexPersistence(namespaces[i]);
                fail("Creates XmlIndexPersistence with bad namespace");
            } catch (PersistenceConfigurationException e) {
                // good
            }
        }
    }

    /**
     * Test of constructor with non-positive file size limit. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_NonPositiveFileSizeLimit()
        throws Exception {
        try {
            new XmlIndexPersistence(0, 0, new File(XmlHelper.PERSISTENCE_DIR_PATH).getCanonicalPath());
            fail("Creates XmlIndexPersistence with non-positive file size limit");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of constructor with negative file size threshold. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_NegativeFileSizeThreshold()
        throws Exception {
        try {
            new XmlIndexPersistence(10, -1, new File(XmlHelper.PERSISTENCE_DIR_PATH).getCanonicalPath());
            fail("Creates XmlIndexPersistence with negative file size threshold");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of constructor with null persistence dir path. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_NullPersistenceDirPath()
        throws Exception {
        try {
            new XmlIndexPersistence(10, 0, null);
            fail("Creates XmlIndexPersistence with null persistence dir path");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of constructor with empty persistence dir path. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_EmptyPersistenceDirPath()
        throws Exception {
        try {
            new XmlIndexPersistence(10, 0, "");
            fail("Creates XmlIndexPersistence with empty persistence dir path");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of constructor with non-absolute persistence dir path. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_NonAbsolutePersistenceDirPath()
        throws Exception {
        try {
            new XmlIndexPersistence(10, 0, XmlHelper.PERSISTENCE_DIR_PATH);
            fail("Creates XmlIndexPersistence with non-absolute persistence dir path");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of constructor with non-directory persistence dir path. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_NonDirectoryPersistenceDirPath()
        throws Exception {
        try {
            new XmlIndexPersistence(10, 0,
                new File(XmlHelper.PERSISTENCE_DIR_PATH, XmlHelper.DOCUMENT_USAGE_FILE).getCanonicalPath());
            fail("Creates XmlIndexPersistence with non-directory persistence dir path");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of constructor with non-existing persistence dir path. Expects IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_NonExistPersistenceDirPath()
        throws Exception {
        try {
            new XmlIndexPersistence(10, 0, new File("non-exist").getCanonicalPath());
            fail("Creates XmlIndexPersistence with non-existing persistence dir path");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of constructor with persistence dir path with non-existing document usage file. Expects
     * IllegalArgumentException.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testConstructor_NonExistDocumentUsageFile()
        throws Exception {
        XmlHelper.cleanupDir();

        try {
            new XmlIndexPersistence(10, 0, new File(XmlHelper.PERSISTENCE_DIR_PATH).getCanonicalPath());
            fail("Creates XmlIndexPersistence with non-existing document usage file");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Test of addDocumentIndex method with non-existing document index. The document index is written to one sequence
     * file. Verifies if the document index was persisted after calling it. This is done by checking the records in
     * the files.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testAddDocumentIndex_NonExistDocumentIndex_Short()
        throws Exception {
        persistence.addDocumentIndex(documentIndex);

        assertTrue("Fails to add document index to files", XmlHelper.verify("add_document_index_short"));
    }

    /**
     * Test of addDocumentIndex method with non-existing document index. The document index is written to multiple
     * sequence files. Verifies if the document index was persisted after calling it. This is done by checking the
     * records in the files.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testAddDocumentIndex_NonExistDocumentIndex_Long()
        throws Exception {
        persistence.addDocumentIndex(TestHelper.createLongDocumentIndex());

        assertTrue("Fails to add document index to files", XmlHelper.verify("add_document_index_long"));
    }

    /**
     * Test of removeDocumentIndex method with unused word source id (use count is zero). The document index is written
     * to one sequence file. Verifies if the document index was removed from persistence after calling it. This is
     * done by checking the records in the files.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testRemoveDocumentIndex_UnusedWordSourceId_Short()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(documentIndex);

        persistence.removeDocumentIndex(wordSourceId);

        assertTrue("Fails to remove document index from files", XmlHelper.verify("remove_document_index_short"));
    }

    /**
     * Test of removeDocumentIndex method with unused word source id (use count is zero). The document index is written
     * to multiple sequence files. Verifies if the document index was removed from persistence after calling it. This
     * is done by checking the records in the files.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testRemoveDocumentIndex_UnusedWordSourceId_Long()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(TestHelper.createLongDocumentIndex());

        persistence.removeDocumentIndex(wordSourceId);

        assertTrue("Fails to remove document index from files", XmlHelper.verify("remove_document_index_long"));
    }

    /**
     * Test of getDocumentIndex method with existing word source id. The document index is written to multiple sequence
     * files. Verifies if the fields of the returned document index were populated.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testGetDocumentIndex_ExistWordSourceId_Long()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(TestHelper.createLongDocumentIndex());

        DocumentIndex documentIndex = persistence.getDocumentIndex(wordSourceId);
        Map words = TestHelper.createLongDocumentIndexWords();

        assertEquals("Returns an incorrect document index", wordSourceId, documentIndex.getWordSourceId());
        assertEquals("Returns an incorrect document index", words, documentIndex.getWords());
    }

    /**
     * Test of addCollectionIndex method with non-existing collection index. Verifies if the collection index was
     * persisted after calling it. This is done by checking the records in the files.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testAddCollectionIndex_NonExistCollectionIndex()
        throws Exception {
        Set wordSourceIds = TestHelper.createCollectionIndexDocumentIds();
        Map words = TestHelper.createDocumentIndexWords();

        // make sure the doucment indexes are persisted
        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            persistence.addDocumentIndex(new DocumentIndex((WordSourceId) i.next(), words));
        }

        persistence.addCollectionIndex(collectionIndex);

        assertTrue("Fails to add collection index to files", XmlHelper.verify("add_collection_index"));
    }

    /**
     * Test of removeCollectionIndex method with existing collection id. Verifies if the collection index was removed
     * from persistence after calling it. This is done by checking the records in the files.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testRemoveCollectionIndex_ExistCollectionId()
        throws Exception {
        Set wordSourceIds = TestHelper.createCollectionIndexDocumentIds();
        Map words = TestHelper.createDocumentIndexWords();

        // make sure the doucment indexes are persisted
        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            persistence.addDocumentIndex(new DocumentIndex((WordSourceId) i.next(), words));
        }

        // make sure the collection index exists
        persistence.addCollectionIndex(collectionIndex);

        persistence.removeCollectionIndex(collectionId);

        assertTrue("Fails to remove collection index from files", XmlHelper.verify("remove_collection_index"));
    }

    /**
     * Test of updateCollectionIndex method with existing collection index. Verifies if the collection index was
     * updated in the persistence after calling it. This is done by checking the records in the files.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testUpdateCollectionIndex_ExistCollectionIndex()
        throws Exception {
        Set wordSourceIds = TestHelper.createCollectionIndexDocumentIds();
        Map words = TestHelper.createDocumentIndexWords();

        // make sure the doucment indexes are persisted
        for (Iterator i = wordSourceIds.iterator(); i.hasNext();) {
            persistence.addDocumentIndex(new DocumentIndex((WordSourceId) i.next(), words));
        }

        // make sure the collection index exists
        persistence.addCollectionIndex(collectionIndex);

        // make sure the new document index is persisted
        persistence.addDocumentIndex(documentIndex);

        wordSourceIds.clear();
        wordSourceIds.add(wordSourceId);
        collectionIndex = new CollectionIndex(null, new HashMap(), wordSourceIds, collectionId);
        persistence.updateCollectionIndex(collectionIndex);

        assertTrue("Fails to update collection index to files", XmlHelper.verify("update_collection_index"));
    }

    /**
     * Test of increaseDocumentUseCount method with existing word source id. Verifies if the use count was updated in
     * the persistence after calling it. This is done by checking the records in the files.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testIncreaseDocumentUseCount_ExistWordSourceId()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(documentIndex);

        persistence.increaseDocumentUseCount(wordSourceId);

        assertTrue("Fails to increase use count in files", XmlHelper.verify("increase_document_use_count"));
    }

    /**
     * Test of decreaseDocumentUseCount method with existing word source id. Verifies if the use count was updated in
     * the persistence after calling it. This is done by checking the records in the files.
     *
     * @throws Exception if any unexpected exception occurs.
     */
    public void testDecreaseDocumentUseCount_ExistWordSourceId()
        throws Exception {
        // make sure the document index exists
        persistence.addDocumentIndex(documentIndex);

        persistence.decreaseDocumentUseCount(wordSourceId);

        assertTrue("Fails to decrease use count in files", XmlHelper.verify("decrease_document_use_count"));
    }
}
